#!/bin/sh
# \
exec wish "$0" "$@"


# Procedure to draw the sensor. The sensor is shown as two cross hairs and a
# red dot.
#
proc drawSensor { canvas x y } {
    global sensorState

    set region [lindex [$canvas configure -scrollregion] end]
    foreach { left top right bottom } $region {	break }

    set xid [$canvas create line $x $top $x $bottom -width 2 \
		 -tag [list sensor_v sensor_line sensor] \
		 -dash { 2 2 }]
    set yid [$canvas create line $left $y $right $y -width 2 \
		 -tag [list sensor_h sensor_line sensor] \
		 -dash { 2 2 }]
    set box [$canvas create rect [expr $x-3] [expr $y-3] \
		 [expr $x+3] [expr $y+3] -width 1 -fill red \
		 -tag [list sensor_box sensor]]

    set sensorState(xpos) $x; set sensorState(xid) $xid
    set sensorState(ypos) $y; set sensorState(yid) $yid
    set sensorState(box) $box
}


# Procedure to draw a machine on the canvas.
#
proc drawMachine { canvas name x y w h text {font "helvetica 14" } } {
    global machineStates

    # A machine is represented by two canvas objects - a word of text and
    # a box surrounding it.
    #
    set objId [$canvas create rect [list 0 0 $w $h] \
		   -tag [list machine $name ${name}_box]\
		   -width 3 -fill green]

    set txtId [$canvas create text [expr $w/2] \
		                   [expr $h/2] -text $text \
		                               -tag [list machine $name]\
		                               -anchor c -font $font]
    $canvas move $name $x $y

    # Set two entries in the global machine names array, so it the user
    # clicks on either the text or the box, we can find what they clicked on
    #
    set machineStates($objId,name) $name
    set machineStates($txtId,name) $name

    set machineStates($name,state) "running"
}


# Procedure to set the colour of the triangular pieces of a valve. The name is
# the valve's tag, as given when it was created. The pieces are 12, 3, 6
# and/or 9, which are the 4 labels given to the 4 pieces which make up a
# valve icon. 
#
proc colourValvePieces { canvas name pieces colour } {
    foreach piece $pieces {
	$canvas itemconfigure ${name}_$piece -fill $colour
    }
}


# Procedure to set the colours of a valve icon. A valve can be moving, open
# or closed. Each state is drawn as a different set of colours.
#
proc setValveColours { canvas name state } {
    global valveStates

    if { $state == "moving" } {
	colourValvePieces $canvas $name {12 3 6 9} "red"
    } elseif { $state == "closed" } {
	colourValvePieces $canvas $name {12 3 6 9} "gray95"
    } else {
	if { $valveStates($name,orientation) == "horizontal" } {
	    colourValvePieces $canvas $name {3 9} "black"
	    colourValvePieces $canvas $name {12 6} "gray95"
	} else {
	    colourValvePieces $canvas $name {3 9} "gray95"
	    colourValvePieces $canvas $name {12 6} "black"
	}
    }
    updatePipes $canvas $name $valveStates($name,state)
}


# Procedure to draw a valve. This looks complicated, but all it actually does
# is draw 4 triangles in the right place, then set an entry in the valveStates.
#
proc drawValve { canvas name x y size orientation state } {
    global valveStates
    
    # A valve is drawn as 4 triangles, which, when put together, would
    # just make a box with an X in it! I reduce the width/height of
    # the triangles by 10% to ensure it looks more like a valve icon
    #
    set offset [expr $size.0/100.0*10]

    # Draw 4 polygons - triangles. I name them after clock face positions:
    # 12 oclock and 6 oclock are the triangles pointing up and down; and 3
    # oclock and 9 oclock are the ones pointing left and right.
    # 
    $canvas create poly 0 $offset \
	                [expr $size/2] [expr $size/2] \
	                0 [expr $size-$offset] \
                        0 $offset \
	-fill gray95 -outline black -width 2 -tag [list valve $name ${name}_9]
    $canvas create poly [expr $size/2] [expr $size/2] \
	                $size $offset \
 	                $size [expr $size-$offset] \
	                [expr $size/2] [expr $size/2] \
	-fill gray95 -outline black -width 2 -tag [list valve $name ${name}_3]

    $canvas create poly $offset 0 \
 	                [expr $size-$offset] 0 \
	                [expr $size/2] [expr $size/2] \
	                $offset 0 \
	-fill gray95 -outline black -width 2 -tag [list valve $name ${name}_12]
    $canvas create poly $offset $size \
	                [expr $size-$offset] $size \
	                [expr $size/2] [expr $size/2] \
                        $offset $size \
	-fill gray95 -outline black -width 2 -tag [list valve $name ${name}_6]

    $canvas move $name $x $y

    # If the user clicks on a valve, but actually hits one of the small gaps
    # between the triangles, nothing happens. Gaps don't count! In order to
    # make behaviour more sensible I wrap an invisible box around the valve
    # drawing. It's this box which actually receives the mouse click.
    #
    foreach { bbox_x1 bbox_y1 bbox_x2 bbox_y2 } [$canvas bbox $name] { break }
    set objId [$canvas create polygon $bbox_x1 $bbox_y1 \
                                      $bbox_x2 $bbox_y1 \
                                      $bbox_x2 $bbox_y2 \
                                      $bbox_x1 $bbox_y2 \
                                      $bbox_x1 $bbox_y1 \
           -tag [list valve_box_$name valve_box] -fill "" -outline "" -width 0]

    set valveStates($objId,name) $name
    set valveStates($name,state) $state
    set valveStates($name,orientation) $orientation
    set valveStates($name,timerId) -1

    setValveColours $canvas $name $state
}


# Procedure to draw a pipe. It's just a rectangle.
#
proc drawPipe { canvas name x y length width orientation } {

    if { $orientation == "horizontal" } {
	$canvas create rect 0 0 $length $width -tag $name -fill black
    } else {
	$canvas create rect 0 0 $width $length -tag $name -fill black
    }

    $canvas move $name $x $y
}


# Procedure to draw the oilfield on the canvas.
#
proc drawImage { canvas } {

    # Since I'm writing this by hand, it's just as easy to code
    # the layout of the image here as it is to put it in an XML file or
    # somewhere else.
    #
    drawMachine $canvas pump         150 320 85 30 "Pump"
    drawMachine $canvas filter       300 320 85 30 "Filter"
    drawMachine $canvas refiner      450 320 85 30 "Refiner"
    drawMachine $canvas sample       545 420 85 30 "Sample"

    drawValve $canvas instreamValve  120 325 20 "horizontal" "open"
    drawValve $canvas infilterValve  270 325 20 "horizontal" "open"
    drawValve $canvas inrefinerValve 420 325 20 "horizontal" "open"
    drawValve $canvas insampleValve  582 380 20 "vertical" "open"
    drawValve $canvas outstreamValve 650 325 20 "horizontal" "open"

    drawPipe $canvas inlet            10 330 110  9 "horizontal"
    drawPipe $canvas inpump          140 330   9 10 "horizontal"
    drawPipe $canvas outpump         236 330  33 10 "horizontal"
    drawPipe $canvas infilter        290 330   9 10 "horizontal"
    drawPipe $canvas outfilter       386 330  33 10 "horizontal"
    drawPipe $canvas inrefiner       440 330   9 10 "horizontal"
    drawPipe $canvas insampler       587 400  20 10 "vertical"
    drawPipe $canvas outlet          670 330 110  9 "horizontal"

    # t piece is drawn specially
    #
    $canvas create polygon 536 330 \
                           649 330 \
                           649 340 \
                           597 340 \
                           597 379 \
                           587 379 \
                           587 340 \
                           536 340 \
                           536 330 -outline black -width 1 \
                                   -tag tpiece -fill black

    $canvas create text 600 10 -anchor nw -justify left -text "" -tag status

    # All the valves have invisible boxes drawn over them. (See the valve
    # procedure.) I raise these boxes so they always get first sight of the
    # mouse clicks.
    #
    foreach valveBox [$canvas find withtag valve_box] {
	$canvas raise $valveBox
    }
    
    # The sensor items get drawn on top of everything
    #
    drawSensor $canvas 400 300
}


# Procedure to update the colours of the drawn pipes. When a pipe has oil
# in it, it gets drawn black. When it hasn't it gets drawn gray. This is
# called whenever a valve is opened or closed.
#
proc updatePipes { canvas valveChanged state } {
    global valveStates

    # This list descibes the pipes and valves. Each line is a 2 element list
    # consisting of a valve name and a sublist. The first element of the 
    # sublist contains the name of the pipe following into the valve; the
    # second element is another list naming the pipes which will contain oil
    # when the valve is open.
    #
    set streamList { instreamValve  {inlet     {inpump outpump}} \
		     infilterValve  {outpump   {infilter outfilter}} \
		     inrefinerValve {outfilter {inrefiner tpiece}} \
		     outstreamValve {tpiece    {outlet}} \
		     insampleValve  {tpiece    {insampler}} }

    # Look at each valve in turn. If the valve isn't closed, and there is oil
    # in the pipe feeding the valve, change the colour of the pipe(s) on the
    # output side of the valve to indicate they have oil in them.
    #
    foreach {valve controlList} $streamList {
	set colour "gray95"
	if { ($valveStates($valve,state) != "closed") &&
	     ([$canvas itemcget [lindex $controlList 0] -fill] == "black") } {
	    set colour "black"
	}

	foreach pipe [lindex $controlList 1] {
	    $canvas itemconfigure $pipe -fill $colour
	}
    }
}


# Procedure called whenever a valve is moved. This is called via a trace so
# that whenever the valveStates array is altered, the valve gets moved on the
# display.
#
# All I do is change the valve state on screen; in a real application this
# would be the code which drives the valve controlling hardware.
#
proc moveValve { canvas arr valveName op } {
    global valveStates

    foreach {valveName subIndex} [split $valveName ","] {break}
    if { $subIndex == "state" } {
	setValveColours $canvas $valveName $valveStates($valveName,state)
    }

    setStatusText $canvas
}


# Procedure called whenever the machine status data is changed. This is called
# via a trace so that whenever the machineStates array is altered, the machine
# gets updated on the display.
#
# All I do is change the machine state on screen; in a real application this
# would be the code which drives the machine controlling hardware.
#
#
proc switchMachine { canvas arr machineName op } {
    global machineStates

    foreach {machineName subIndex} [split $machineName ","] {break}

    # I'm only interested in the state entries in the array
    #
    if { $subIndex == "state" } {
	$canvas itemconfigure ${machineName}_box -fill \
	    [expr { $machineStates($machineName,state) == "running" ? \
			                                "green" : "slategray"}]
    }

    setStatusText $canvas
}


# Procedure called when the user clicks a valve icon on the screen.
#
proc operateValve { canvas xpos ypos } {
    global valveStates

    # Find the valve they clicked on. They might have clicedk right on the
    # border of a valve and pipe, so if the valve is not found I exit quietly
    # 
    set itemId [$canvas find closest $xpos $ypos]
    if { [catch { set valveName $valveStates($itemId,name) }] } { return }

    # Once a valve is set moving ignore all other commands until it's done
    #
    if { $valveStates($valveName,timerId) != -1 } {
	return
    }
    
    # Decide which way the valve is going to go
    # 
    set targetState \
	[expr {$valveStates($valveName,state) == "open" ? "closed" : "open"}]

    # In the absence of any real valve hardware, I just use a timer to set
    # the desired valve state in a few seconds time.
    #
    set valveStates($valveName,state) "moving"
    set valveStates($valveName,timerId) \
	[after 5000 "set valveStates($valveName,timerId) -1; \
                     set valveStates($valveName,state) $targetState"]
}


# Procedure called when the user clicks a machine icon on the screen.
#
proc operateMachine { canvas xpos ypos } {
    global machineStates

    # Find the machine they clicked on. They might have clicked right on the
    # border of a machine and pipe, so if the machine is not found I exit
    # quietly
    # 
    set itemId [$canvas find closest $xpos $ypos]
    if { [catch { set machineName $machineStates($itemId,name) }] } { return }
    
    # Toggle the state of the machine. This triggers the code which actually
    # drives the machine hardware.
    #
    set machineStates($machineName,state) \
	[expr {($machineStates($machineName,state) == "running") ? "stopped" :\
		                                                   "running"}]
}


# Procedure to update the text displaying the sensor information.
#
proc setSensorText { args } {
    global sensorState

    set sensorState(text) [format "Sensor at %03d,%03d reading %05.2f" \
				  [expr int($sensorState(xpos))] \
			          [expr int($sensorState(ypos))] \
			                $sensorState(value)]
}


# Procedure to update the status text appearing in the top left corner
#
proc setStatusText { canvas } {
    global machineStates valveStates

    set statusText ""
    foreach valve { instreamValve infilterValve inrefinerValve \
	            insampleValve outstreamValve } {
	append statusText \
	    [string totitle "$valve\t: $valveStates($valve,state)\n"]
    }
    append statusText "\n"
    foreach machine { pump filter refiner sample } {
	append statusText \
	    [string totitle "$machine\t\t: $machineStates($machine,state)\n"]
    }

    $canvas itemconfigure status -text $statusText
}


# Procedure called whenever the sensor is moved onscreen. All this code does
# is change the position of the sensor widgets onscreen. In a real application
# it would need to control the sensor moving hardware too.
#
proc moveSensor { canvas position axis } {
    global sensorState

    set region [lindex [$canvas configure -scrollregion] end]
    foreach { left top right bottom } $region {	break }

    if { $position == "centre" } {
	# Centre the specified axis
	#
	set position [expr { $axis == "v" ? [expr {($right-$left)/2}] : \
				            [expr {($bottom-$top)/2}] }]
    }

    # Change the coords of the specified axis to the requested place
    # 
    if { $axis == "v" } {
	if { $position < $left }  { set position $left }
	if { $position > $right } { set position $right }
	set sensorState(xpos) $position
	$canvas coords sensor_v $position $top $position $bottom
    } else {
	if { $position < $top    } { set position $top }
	if { $position > $bottom } { set position $bottom }
	set sensorState(ypos) $position
	$canvas coords sensor_h $left $position $right $position
    }

    # Move the box to the middle of the cross hairs
    #
    $canvas coords sensor_box \
	[expr $sensorState(xpos)-3] [expr $sensorState(ypos)-3] \
	[expr $sensorState(xpos)+3] [expr $sensorState(ypos)+3]
}


# Procedure to handle the mouse interaction with the widgets which represent
# the sensor.
#
proc operateSensor { canvas action {xpos 0} {ypos 0} } {
    global sensorState

    switch $action {
	"enter" {
	    # The mouse has entered one of the lines, so change cursor shape
	    #
	    if { [$canvas find withtag current] == $sensorState(xid) } {
		$canvas config -cursor sb_h_double_arrow
	    } else {
		$canvas config -cursor sb_v_double_arrow
	    }
	}

	"leave" {
	    # The mouse has left the lines, so put the cursor shape back
	    #
	    $canvas config -cursor left_ptr
	}

	"move" {
	    # Something is being dragged. Find what is under the mouse and move
	    # it to the mouse location
	    #
	    if { [$canvas find withtag current] == $sensorState(xid) } {
		moveSensor $canvas $xpos v
	    } elseif { [$canvas find withtag current] == $sensorState(yid) } {
		moveSensor $canvas $ypos h
	    } elseif { [$canvas find withtag current] == $sensorState(box) } {
		moveSensor $canvas $xpos v
		moveSensor $canvas $ypos h
	    }
	}

	"centre" {
	    # The sensor should be placed in the centre of the display
	    #
	    moveSensor $canvas "centre" v
	    moveSensor $canvas "centre" h
	}
    }
}



#
# Main program starts here
#
set valveStates(instreamValve,state)  "closed"
set valveStates(infilterValve,state)  "closed"
set valveStates(inrefinerValve,state) "closed"
set valveStates(insampleValve,state)  "closed"
set valveStates(outstreamValve,state) "closed"

set machineStates(pump,state)         "stopped"
set machineStates(filter,state)       "stopped"
set machineStates(refiner,state)      "stopped"
set machineStates(sample,state)       "stopped"

set sensorState(xpos)                 0
set sensorState(ypos)                 0
set sensorState(value)                0
set sensorState(text)                 ""


# Draw the GUI
#
wm resizable . 0 0
set canvas [canvas .canvas -width 800 -height 600 \
                           -scrollregion {0 0 800 600} \
                           -highlightthickness 0 \
                           -borderwidth 0 \
                           -background gray95]
pack $canvas -side top -fill both -expand true
drawImage $canvas

set sensorFrame [frame $canvas.s -relief groove -bd 2]
button $sensorFrame.c -text "Centre sensor" \
    -command [list operateSensor $canvas "centre"]
label $sensorFrame.l -textvariable sensorState(text)
pack $sensorFrame.c $sensorFrame.l -side left -padx 5 -pady 4
$canvas create window 5 5 -anchor nw -window $sensorFrame


# Bind button events on a valve, machine or sensor to the right subroutines
#
$canvas bind valve_box <Button-1> { operateValve   %W [%W canvasx %x] \
					              [%W canvasy %y] }

$canvas bind machine <Button-1>   { operateMachine %W [%W canvasx %x] \
					              [%W canvasy %y] }

$canvas bind sensor_line <Enter>  { operateSensor %W enter }
$canvas bind sensor_line <Leave>  { operateSensor %W leave }
$canvas bind sensor <B1-Motion>   { operateSensor %W move [%W canvasx %x] \
					                  [%W canvasy %y] }

# When the sensor, valve or machine variables are changed, call the routines
# which ensure the display is updated.
#
# Note that I'm using the old, deprecated syntax here. Some Linux distros still
# ship Tcl/Tk 8.3 which doesn't support the new syntax.
#
trace variable sensorState   w setSensorText
trace variable valveStates   w [list moveValve $canvas]
trace variable machineStates w [list switchMachine $canvas]

# Set a loop running which resamples the temperature at the sensor every 3
# seconds. For the sake of this demonstration I think of a random number; in
# a real application the sensor hardware would need to be examined.
#
proc changeTemperature {} {
    set ::sensorState(value) [expr {rand() * 100}]
    after 3000 changeTemperature
}
changeTemperature

setStatusText $canvas